;
; Platonic Solids demo program - S. Yockey, April 2023
;
; 3d rotation code was adapted from a Simons' Basic program from 
; https://retro64.altervista.org/blog/3d-graphics-rotating-cube-with-freebasic-pc-and-simons-basic-c64/
; that I originally coded in Slang for the DTV64. The Prog8 package includes a spinning cube demo that
; does not require floating point, but since we are doing all of the rotation calculations prior to any
; plotting, this is a minor factor.
;
; Version 2 - added back face culling (of a sort) to all objects.
;             This removal of hidden lines is 'quick and dirty' and 'tuned' to the given parameters.
;			  It is not a generalized procedure and requires 'tuning' for each object.
;			  Also did some optimizing to reduce code size.
;
; This program is compatible with x16emu & Box16 - R41 & R42

%import syslib
%import floats
%import textio

main { 
    uword[61] vx 
    uword[61] vy
    word[61] vz
    
    word[] vector1 = [0,0,0]
    word[] vector2 = [0,0,0]
    word[] vector3 = [0,0,0]
    
    ubyte np
    bool hidden = true
    
    float tx
    float ty
    float tz
    
    uword tt
    
    word f1
	word f2
	word f3
	word f4
	word f5
	word f6
	word f7
	word f8
	word f9
	word f10
	word f11
	word f12
	word f13
	word f14
	word f15
	word f16
	word f17
	word f18
	word f19
	word f20
	word f21
	word f22
	word f23
	word f24
	word f25
	word f26
	word f27
	word f28
	word f29
	word f30
	word f31
	word f32
	
	word[4] t_faces
    word[4] b_faces
    
 
; flags for each type, so we only calculate point rotations 1 time
	bool c_f = false
	bool d_f = false
	bool o_f = false
	bool i_f = false
	bool b_f = false
		
; master flag for whether or not to calculate 3d transformation
	bool calculate = true	
       
    const float PHI = 1.618		;close enough
    const ubyte TETRAHEDRON = 1
    const ubyte OCTAHEDRON = 4
    const ubyte CUBE = 7
    const ubyte DODECAHEDRON = 10
    const ubyte ICOSAHEDRON = 13
    const ubyte BUCKYBALL = 16
    
; tweaked values to keep objects within the 'screen refresh' area
    const float L = 19
    const float L1 = 7
    
    float[] icosx = [0,0,0,0,L,-L,-L,L,PHI*L,PHI*L,-PHI*L,-PHI*L]
    float[] icosy = [-L,L,L,-L,-PHI*L,-PHI*L,PHI*L,PHI*L,0,0,0,0]
    float[] icosz = [PHI*L,PHI*L,-PHI*L,-PHI*L,0,0,0,0,L,-L,-L,L]
    
    float[] dodecx = [-L,-L,-L,-L, L,L, L, L,-L/PHI,L/PHI,-L/PHI, L/PHI, 0    , 0    ,0    , 0    ,-L*PHI,-L*PHI, L*PHI,L*PHI]
    float[] dodecy = [-L, L, L,-L,-L,L,-L, L, 0    ,0    , 0    , 0    ,-L*PHI,-L*PHI,L*PHI, L*PHI,-L/PHI, L/PHI,-L/PHI,L/PHI]
    float[] dodecz = [ L, L,-L,-L, L,L,-L,-L, L*PHI,L*PHI,-L*PHI,-L*PHI, L/PHI,-L/PHI,L/PHI,-L/PHI,0     , 0    , 0    ,0]
    
    float[] octx = [-L,0,L, 0,0, 0]
    float[] octy = [ 0,L,0,-L,0, 0]
    float[] octz = [ 0,0,0, 0,L,-L]
    
    float[] cubex = [L,L,L,L,-L,-L,-L,-L]
    float[] cubey = [-L,L,L,-L,-L,L,L,-L]
    float[] cubez = [-L,-L,L,L,L,L,-L,-L]
    
; buckyball breaks Prog8 limitation on array size!!!!! Workaround required    
    float[] buckx = [0,0,0,0,
					L1, L1, -L1, -L1,
					L1*3*PHI, L1*3*PHI, -L1*3*PHI, -L1*3*PHI,
					L1*2, L1*2, L1*2, L1*2,
					-L1*2, -L1*2, -L1*2, -L1*2,
					L1*(2*PHI+1), L1*(2*PHI+1), L1*(2*PHI+1), L1*(2*PHI+1),
					-L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1),
					L1*PHI, L1*PHI, L1*PHI, L1*PHI, 
					-L1*PHI,-L1*PHI, -L1*PHI, -L1*PHI, 	
					L1, L1, L1, L1, 
					-L1,-L1, -L1, -L1,	
					L1*(2+PHI), L1*(2+PHI), L1*(2+PHI),	L1*(2+PHI),
					 -L1*(2+PHI),-L1*(2+PHI)]
															
	float[] buckx2 = [-L1*(2+PHI), -L1*(2+PHI), L1*2*PHI, L1*2*PHI, L1*2*PHI, L1*2*PHI, -L1*2*PHI, -L1*2*PHI, -L1*2*PHI,-L1*2*PHI]		
										
					
    float[] bucky = [L1,L1,-L1,-L1,
    				L1*3*PHI, -L1*3*PHI, L1*3*PHI, -L1*3*PHI,
    				0,0,0,0,
    				L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1),
    				L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1),
    				L1*PHI, L1*PHI, -L1*PHI, -L1*PHI,
    				L1*PHI, L1*PHI, -L1*PHI, -L1*PHI,
    				L1*2, L1*2, -L1*2, -L1*2, 
    				L1*2, L1*2, -L1*2, -L1*2,
    				L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), -L1*(2+PHI),
    				L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), -L1*(2+PHI),    
    				L1*2*PHI, L1*2*PHI, -L1*2*PHI, -L1*2*PHI,
    				L1*2*PHI, L1*2*PHI ]    				
    				    				
    float[] bucky2 = [-L1*2*PHI, -L1*2*PHI, L1, L1, -L1, -L1, L1, L1, -L1, -L1]
    				
    				
    float[] buckz = [L1*3*PHI, -L1*3*PHI, L1*3*PHI, -L1*3*PHI,
    				0,0,0,0,
    				L1, -L1, L1, -L1,
    				L1*PHI, -L1*PHI, L1*PHI, -L1*PHI,
    				L1*PHI, -L1*PHI, L1*PHI, -L1*PHI,
    				L1*2, -L1*2, L1*2, -L1*2,
    				L1*2, -L1*2, L1*2, -L1*2,
    				L1*(2*PHI+1), -L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1),
    				L1*(2*PHI+1), -L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1),    
    				L1*2*PHI, -L1*2*PHI, L1*2*PHI, -L1*2*PHI, 
    				L1*2*PHI, -L1*2*PHI, L1*2*PHI, -L1*2*PHI,
    				L1, -L1, L1, -L1, 
    				L1, -L1]
    				
    float[] buckz2 = [ L1,-L1, L1*(2+PHI), -L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI)]	
    				

	sub start() {
		float c
		float s
		float xt
		float yt
		float zt
		float t
		float t1
		float s1
		float c1
		float[50] x
		float[10] x2
		float[50] y		
		float[10] y2
		float[50] z		
		float[10] z2		
				
		
		word fs
		float rx
		uword duration
		bool debug = false
						
		ubyte key = 0
		ubyte q
		ubyte body_type = OCTAHEDRON
		ubyte numpts
			
		fs = 200
		rx = -0.1
		
		repeat {
		
			void cx16.screen_mode(3, false) 	;40x25textmode
			txt.color2(5,0)
			txt.clear_screen()
			txt.plot(10,1)
			txt.print("the platonic solids")
			txt.plot(5,9)
    		txt.print("c) cube")
    		txt.plot(5,11)
    		txt.print("d) dodecahedron")
    		txt.plot(5,13)
    		txt.print("o) octahedron") 
    		txt.plot(5,15)
    		txt.print("b) buckyball")
			txt.plot(5, 17)
    		txt.print("i) icosahedron")
    		txt.plot(5,19)
    		txt.print("q) quit")    	
    	
    		while key != 'q' {
    			key = c64.GETIN()
    				
    			when key {
    				'q' -> { 
    					goto quit }
    					
    				'c' -> {
    					body_type = CUBE
    					break
    				}
    						
    				'd' -> {
    					body_type = DODECAHEDRON
    					break
    				}
    					 
    				'o' -> {
    					body_type = OCTAHEDRON
    					break
    				}
    					 
    				'i' -> {
    					body_type = ICOSAHEDRON
    					break
    				}
    				'b' -> {
    					body_type = BUCKYBALL
    					break
    				}    			
    			}
    		}
    						
    		when body_type {
    			DODECAHEDRON -> {
    				numpts = 20
    				cx16.memory_copy(&dodecx,&x,100)
    				cx16.memory_copy(&dodecy,&y,100)
					cx16.memory_copy(&dodecz,&z,100) } 
				
    			OCTAHEDRON -> {
    				numpts = 6
    				octy[1] = 30
    				octy[3] = -30
    				cx16.memory_copy(&octx,&x,40)
    				cx16.memory_copy(&octy,&y,40)
					cx16.memory_copy(&octz,&z,40) }				
						
    			CUBE -> {
    				numpts = 8
    				cx16.memory_copy(&cubex,&x,40)
    				cx16.memory_copy(&cubey,&y,40)
					cx16.memory_copy(&cubez,&z,40) }							
										
    			ICOSAHEDRON -> {
    				numpts = 12
    				cx16.memory_copy(&icosx,&x,60)
    				cx16.memory_copy(&icosy,&y,60)
					cx16.memory_copy(&icosz,&z,60) }										
														
    			BUCKYBALL -> {
    				numpts = 60			
    				cx16.memory_copy(&buckx,&x,250)
    				cx16.memory_copy(&bucky,&y,250)
					cx16.memory_copy(&buckz,&z,250) 
					cx16.memory_copy(&buckx2,&x+250,50)
    				cx16.memory_copy(&bucky2,&y+250,50)
					cx16.memory_copy(&buckz2,&z+250,50) } 														
						
			}
				
			txt.clear_screen()			
			txt.plot(1, 1)
			when body_type {
				BUCKYBALL -> {
					txt.print("buckyball - aka truncated icosahedron")
					txt.plot(5,3)
					txt.print("60 vertices")
					txt.plot(5,4)
					txt.print("90 edges")
					txt.plot(5,5)
					txt.print("32 faces")
					if not b_f {			
						txt.plot(1,10)
    					txt.print("err...the oracles refuse to hear us,") 
    					txt.plot(1,12)
    					txt.print("since the buckyball, aka truncated")				
    					txt.plot(1,14)
    					txt.print("icosahedron, is not a platonic solid.")
    					txt.plot(5,18)
    					txt.print("...this could take a while...") 
    					calculate = true }	
    				else {
    					calculate = false }
    			}
    			
    			CUBE -> {
					txt.print("cube - fancy name unknown")
					txt.plot(5,3)
					txt.print("8 vertices")
					txt.plot(5,4)
					txt.print("12 edges")
					txt.plot(5,5)
					txt.print("6 faces")    	
			
					if not c_f {			
    					calculate = true }	
    				else {
    					calculate = false }
					}  
			
    			DODECAHEDRON -> {
					txt.print("dodecahedron")
					txt.plot(5,3)
					txt.print("20 vertices")
					txt.plot(5,4)
					txt.print("30 edges")
					txt.plot(5,5)
					txt.print("12 faces") 
			 
					if not d_f {			
    					calculate = true }	
    				else {
    					calculate = false }			
			  	
				}  			
					
    			ICOSAHEDRON -> {
					txt.print("icosahedron")
					txt.plot(5,3)
					txt.print("12 vertices")
					txt.plot(5,4)
					txt.print("30 edges")
					txt.plot(5,5)
					txt.print("20 faces") 
			  
					if not i_f {			
    					calculate = true }	
    				else {
    					calculate = false }			
			 	
				}  					
			 	
    			OCTAHEDRON -> {
					txt.print("octahedron - aka diamond or rhombus")
					txt.plot(5,3)
					txt.print("6 vertices")
					txt.plot(5,4)
					txt.print("12 edges")
					txt.plot(5,5)
					txt.print("8 faces")  
			
					if not o_f {			
    					calculate = true }	
    				else {
    					calculate = false }			  	
					}   				
				
    	}
    
    	if body_type != BUCKYBALL and calculate {
    		txt.plot(1,10)
    		txt.print("please wait....") 
    		txt.plot(1,14)
    		txt.print("while we consult the platonic oracles.")				
 		}
    	
    	txt.plot(5,25)
    	txt.print("press 'space' to pause rotation") 
    	txt.plot(5,26)
    	txt.print(", 'h' to toggle hidden lines") 
    	txt.plot(5,27)
    	txt.print("or 'm' to return to the menu") 
    	
    	if  not calculate {
    		sys.wait(180)
    		goto nocalc }
    	
; calculate 3d transformation
		for q in 0 to 64 {
	
			c = floats.cos(rx)
			s = floats.sin(rx)	
		
			for np in 1 to numpts {
				if np <= 50 {
					tx = x[np-1]
					ty = y[np-1]
					tz = z[np-1] }
				else {
					tx = x2[np-51]
					ty = y2[np-51]
					tz = z2[np-51]	}				
							
												
			yt = ty 
			ty = c*yt-(s*tz)
			tz = s*yt+(c*tz)	
			xt = tx
			tx = c*xt+(s*tz)	
			tz = -s*xt+(c*tz)
			vz[np] = (tz+50) as word			
			
			
			xt = tx
			tx = xt*c-(ty*s)
			ty = xt*s+(ty*c)	
			t = tx*fs
			t1 = tz+fs		
			t1 = t/t1			
			t = t1+160		
			vx[np] = t as uword				
							
			t = ty*fs
			t1 = tz+fs
			t1 = t/t1
			t = t1+100
			vy[np] = t as uword
		
		 
		}

	;copy calculated points to banked ram
	
	cx16.rambank(body_type)
	cx16.memory_copy(&vx,$a000 + (q as uword * 121 as uword),121)	
	cx16.rambank(0)

	cx16.rambank(body_type+1)
	cx16.memory_copy(&vy,$a000 + (q as uword * 121 as uword),121)	
	cx16.rambank(0)	
	
	cx16.rambank(body_type+2)
	cx16.memory_copy(&vz,$a000 + (q as uword * 121 as uword),121)	
	cx16.rambank(0)	

	rx += 0.098     
	}
	
nocalc:


	
	;320x200x256
	gfx3.screen_mode(2)
	
	;clears screen and back buffer	
	init_buffers()
	
	q = 0
	while key != 'm' {
		c64.SETTIM(0, 0, 0)

		; restore saved points from banked ram to working arrays		
		cx16.rambank(body_type)
		cx16.memory_copy($a000 + (q as uword * 121 as uword),&vx,121)
		cx16.rambank(0)		
	
		cx16.rambank(body_type+1)
		cx16.memory_copy($a000 + (q as uword * 121 as uword),&vy,121)
		cx16.rambank(0)			
				
		cx16.rambank(body_type+2)
		cx16.memory_copy($a000 + (q as uword * 121 as uword),&vz,121)
		cx16.rambank(0)			

    	;clear drawing area only
    	gfx3.fillrect(121,261,76,76,0)
           
		when body_type {
			DODECAHEDRON -> draw_dodec()	
				
			OCTAHEDRON -> draw_oct()	
			
			CUBE -> draw_cube()
			
			ICOSAHEDRON -> draw_icosa()
				
			BUCKYBALL -> draw_bucky()
		}				
			    
		;copy onto visible screen
		sys.waitvsync()
		blit2(120,260,76,76)
			
		duration = c64.RDTIM16()
		key = c64.GETIN()
		
		if key == ' ' {
			getkey() }
			
		if key == 'h' {
			hidden = not hidden }
			
		if q < 64 q += 1
		else q = 0
			
			
	} 
	
	;retore text mode
	gfx3.screen_mode(0)
	void cx16.screen_mode(3, false) 	;40x25textmode		
	
	txt.color2(1,0)	
	txt.clear_screen()
    txt.plot(5,12)
    floats.print_f(60 as float/duration as float)
    txt.print(" frames per second!\n")        
    txt.plot(5,14)
    txt.print("press any key to continue...")
    
    getkey()
    key = 0
    }
    
quit:
	
} 

sub draw_cube() {
	ubyte[] line_pairs = [1,2,2,3,3,4,4,1,4,5,5,6,6,3,6,7,7,8,8,5,2,7,8,1]	;entire object
	; object by face
	ubyte[] f1_pairs = [1,2,2,3,3,4,4,1]
	ubyte[] f2_pairs = [3,4,4,5,5,6,6,3]
	ubyte[] f3_pairs = [5,6,6,7,7,8,8,5]
	ubyte[] f4_pairs = [1,2,2,7,7,8,8,1]
	ubyte[] f5_pairs = [2,3,3,6,6,7,7,2]
	ubyte[] f6_pairs = [1,4,4,5,5,8,8,1]
	word[6] faces

	if hidden {
 
    	faces[0] = (vz[1] + vz[2] + vz[3] + vz[4])/4
    	f1 = faces[0]
    	faces[1] = (vz[3] + vz[4] + vz[5] + vz[6])/4
    	f2 = faces[1]
    	faces[2] = (vz[5] + vz[6] + vz[7] + vz[8])/4
    	f3 = faces[2]
    	faces[3] = (vz[1] + vz[2] + vz[7] + vz[8])/4
    	f4 = faces[3]
    	faces[4] = (vz[2] + vz[3] + vz[6] + vz[7])/4
    	f5 = faces[4]
    	faces[5] = (vz[1] + vz[4] + vz[5] + vz[8])/4
    	f6 = faces[5]
    
   		sort(faces)
   	
	
		if f1 <= $30 {	;face1
			draw_object(&f1_pairs,3) 
		}
		if f2 <= $30 {	;face2
			draw_object(&f2_pairs,3) 
		}
		if f3 <= $30 {	;face3
			draw_object(&f3_pairs,3)		 
		}				
		if f4 <= $30 {	;face4
			draw_object(&f4_pairs,3)	
		}			
		if f5 <= $30 {	;face5
			draw_object(&f5_pairs,3)
		}		
		if f6 <= $30 {	;face6
			draw_object(&f6_pairs,3)
		}
	
	}
	else {
		draw_object(&line_pairs,11)	
	}
	
	c_f = true			
}

sub draw_oct() {
    ; drawing to back buffer (offscreen)
    ubyte[] line_pairs = [1,2,2,3,3,4,4,1,1,5,1,6,2,5,2,6,3,5,3,6,4,5,4,6]
    ubyte[] f1_pairs = [1,2,2,5,5,1]
    ubyte[] f2_pairs = [5,2,2,3,3,5]
    ubyte[] f3_pairs = [3,2,2,6,6,3]
    ubyte[] f4_pairs = [6,2,2,1,1,6]
    ubyte[] f5_pairs = [1,4,4,5,5,1]
    ubyte[] f6_pairs = [5,4,4,3,3,5]
	ubyte[] f7_pairs = [3,4,4,6,6,3]
	ubyte[] f8_pairs = [6,4,4,1,1,6]
	
	if hidden {
    	t_faces[0] = (vz[1] + vz[2] + vz[5]);/3
    	f1 = t_faces[0]
    	t_faces[1] = (vz[2] + vz[3] + vz[5]);/3
    	f2 = t_faces[1]
    	t_faces[2] = (vz[2] + vz[3] + vz[6]);/3
    	f3 = t_faces[2]
    	t_faces[3] = (vz[1] + vz[2] + vz[6]);/3
    	f4 = t_faces[3]
    
    	b_faces[0] = (vz[1] + vz[4] + vz[5]);/3
    	f5 = b_faces[0]
    	b_faces[1] = (vz[3] + vz[4] + vz[5]);/3
    	f6 = b_faces[1]
    	b_faces[2] = (vz[3] + vz[4] + vz[6]);/3
    	f7 = b_faces[2]
    	b_faces[3] = (vz[1] + vz[4] + vz[6]);/3
    	f8 = b_faces[3]
    
    	sort(t_faces)
    	sort(b_faces)
    
		;vector(1,6,&vector1)
	
	
    
		if f1 <= t_faces[1] and f1 < $9f  {	;face1
			draw_object(&f1_pairs,2)
			
 
		}
	
		if f2 <= t_faces[1] and f2 < $9f {	;face2
			draw_object(&f2_pairs,2)
			

		}
	
		if f3 <= t_faces[1] and f3 < $9f {	;face3
			draw_object(&f3_pairs,2)
			
 
		}
	
		if f4 <= t_faces[1] and f4 < $9f {	;face4
			draw_object(&f4_pairs,2)
			

		}   
	
		if f5 <= b_faces[1] and f5 < $9f {	;face5
			draw_object(&f5_pairs,2)
			

		}  
	
		if f6 <= b_faces[1] and f6 < $9f {	;face6
			draw_object(&f6_pairs,2)
			

		}  	
	
		if f7 <= b_faces[1] and f7 < $9f {	;face7
			draw_object(&f7_pairs,2)
			

		}  	
	
		if f8 <= b_faces[1]  and f8 < $9f{	;face8
			draw_object(&f8_pairs,2)
			

		}  	
	
	}
	else {
    	draw_object(&line_pairs,11)
	}
	
	o_f = true		
}



sub draw_dodec() {
    ; drawing to back buffer (offscreen)
    word[12] faces
    ubyte[] line_pairs =	[8,16,14,7,3,16,19,5,5,13,
							13,1,1,9,9,10,10,5,10,6,
							6,20,20,19,19,7,8,12,12,7,
							12,11,11,4,4,14,4,17,17,1,
							17,18,18,2,2,15,15,6,15,16,
							20,8,3,18,3,11,2,9,13,14]	
								
	ubyte[] f1_pairs = [1,9,9,10,10,5,5,13,13,1]							
    ubyte[] f2_pairs = [1,13,13,14,14,4,4,17,17,1]
    ubyte[] f3_pairs = [14,7,7,12,12,11,11,4,4,14]
    ubyte[] f4_pairs = [7,19,19,20,20,8,8,12,12,7]
    ubyte[] f5_pairs = [19,5,5,10,10,6,6,20,20,19]
    ubyte[] f6_pairs = [13,5,5,19,19,7,7,14,14,13]
    
    ubyte[] f7_pairs = [8,16,16,15,15,6,6,20,20,8]
    ubyte[] f8_pairs = [6,15,15,2,2,9,9,10,10,6]
    ubyte[] f9_pairs = [9,1,1,17,17,18,18,2,2,9]
    ubyte[] f10_pairs = [18,17,17,4,4,11,11,3,3,18]
    ubyte[] f11_pairs = [18,2,2,15,15,16,16,3,3,18]
    ubyte[] f12_pairs = [3,11,11,12,12,8,8,16,16,3]
    
	if hidden {
    
	    faces[0] = (vz[1] + vz[9] + vz[10] + vz[5] + vz[13])/4
		f1 = faces[0]
	    faces[1] = (vz[1] + vz[13] + vz[14] + vz[4] + vz[17])/4
    	f2 = faces[1]
    	faces[2] = (vz[14] + vz[7] + vz[12] + vz[11] + vz[4])/4
    	f3 = faces[2]
    	faces[3] = (vz[7] + vz[19] + vz[20] + vz[8] + vz[12])/4
    	f4 = faces[3]
    	faces[4] = (vz[19] + vz[5] + vz[10] + vz[6] + vz[20])/4
    	f5 = faces[4]
    	faces[5] = (vz[13] + vz[5] + vz[19] + vz[7] + vz[14])/4
    	f6 = faces[5]
    
    	faces[6] = (vz[8] + vz[16] + vz[15] + vz[6] + vz[20])/4
    	f7 = faces[6]
    	faces[7] = (vz[6] + vz[15] + vz[2] + vz[9] + vz[10])/4
    	f8 = faces[7]
    	faces[8] = (vz[9] + vz[1] + vz[17] + vz[18] + vz[2])/4
    	f9 = faces[8]
    	faces[9] = (vz[18] + vz[17] + vz[4] + vz[11] + vz[13])/4
    	f10 = faces[9]
    	faces[10] = (vz[18] + vz[2] + vz[15] + vz[16] + vz[3])/4
    	f11 = faces[10]
    	faces[11] = (vz[3] + vz[11] + vz[12] + vz[8] + vz[16])/4
    	f12 = faces[11]
    
   		sort(faces)
   	
    
    	;face 1
    	if f1 <= $3a {
    		draw_object(&f1_pairs,4)
    	}

		;face 2
		if f2 <= $3a {
			draw_object(&f2_pairs,4)	
		 }

		;face 3
		if f3 <= $3a {
			draw_object(&f3_pairs,4)	
		 }

		;face 4			
		if f4 <= $3a  {		
			draw_object(&f4_pairs,4)	
 		}			
				
		;face 5	
		if f5 <= $3a {
			draw_object(&f5_pairs,4)
		}		
				
		;face 6			
		if f6 <= $3a {		
			draw_object(&f6_pairs,4)
		}		
				
		;face 7			
		if f7 <= $3a {		
			draw_object(&f7_pairs,4)
 		}		
				
		;face 8			
		if f8 <= $3a  {		
			draw_object(&f8_pairs,4)	
		}		
				
		;face 9			
		if f9 <= $3a  {		
			draw_object(&f9_pairs,4)
		}		
				
		;face 10			
		if f10 <= $3a  {		
			draw_object(&f10_pairs,4)	
		}			
				
		;face 11			
		if f11 <= $3a {			
			draw_object(&f11_pairs,4)	
 		}			
				
		;face 12			
		if f12 <= $3a {		
			draw_object(&f12_pairs,4)	
 		}			
				
	}			
	else {							
		draw_object(&line_pairs,29)			
	}
	d_f = true		
}

sub draw_icosa() {
    ; drawing to back buffer (offscreen)
    word[20] faces
    ubyte[] line_pairs =	[1,12,12,2,2,1,
    						1,6,6,5,5,1,
    						1,9,9,2,2,8,
    						8,9,9,10,10,8,
    						8,3,3,7,7,8,
    						7,2,7,11,11,6,
    						6,12,6,4,4,5,
    						5,10,10,4,4,3,			
    						9,5,12,11,12,7,
    						11,4,11,3,10,3]
    						
    ubyte[] f1_pairs = [1,9,9,5,5,1]							
    ubyte[] f2_pairs = [1,12,12,6,6,1]
    ubyte[] f3_pairs = [1,12,12,2,2,1]
    ubyte[] f4_pairs = [1,2,2,9,9,1]
    ubyte[] f5_pairs = [1,5,5,6,6,1]
    
    ubyte[] f6_pairs = [2,9,9,8,8,2]
    ubyte[] f7_pairs = [2,8,8,7,7,2]
    ubyte[] f8_pairs = [2,7,7,12,12,2]
    
    ubyte[] f9_pairs = [9,8,8,10,10,9]
    ubyte[] f10_pairs = [9,10,10,5,5,9]
    ubyte[] f11_pairs = [10,5,5,4,4,10]
    ubyte[] f12_pairs = [5,4,4,6,6,5]						
    
    ubyte[] f13_pairs = [6,11,11,12,12,6]							
    ubyte[] f14_pairs = [12,11,11,7,7,12]
    ubyte[] f15_pairs = [8,3,3,7,7,8]
    ubyte[] f16_pairs = [10,3,3,8,8,10]
    ubyte[] f17_pairs = [3,7,7,11,11,3]
    
    ubyte[] f18_pairs = [4,3,3,10,10,4]
    ubyte[] f19_pairs = [4,11,11,3,3,4]
    ubyte[] f20_pairs = [6,11,11,4,4,6]
    
    if hidden {
    
     	faces[0] = (vz[1] + vz[9] + vz[5])
		f1 = faces[0]
	    faces[1] = (vz[1] + vz[12] + vz[6])
    	f2 = faces[1]
    	faces[2] = (vz[1] + vz[12] + vz[2])
    	f3 = faces[2]
    	faces[3] = (vz[1] + vz[2] + vz[9])
    	f4 = faces[3]
    	faces[4] = (vz[1] + vz[5] + vz[6])
    	f5 = faces[4]
    	faces[5] = (vz[2] + vz[9] + vz[8])
    	f6 = faces[5]
    	faces[6] = (vz[2] + vz[8] + vz[7])
    	f7 = faces[6]
    	faces[7] = (vz[2] + vz[7] + vz[12])
    	f8 = faces[7]
    	faces[8] = (vz[9] + vz[8] + vz[10])
    	f9 = faces[8]
    	faces[9] = (vz[9] + vz[10] + vz[5])
    	f10 = faces[9]
    	faces[10] = (vz[10] + vz[5] + vz[4])
    	f11 = faces[10]
    	faces[11] = (vz[5] + vz[4] + vz[6])
    	f12 = faces[11]
    	
    	
    	faces[12] = (vz[6] + vz[11] + vz[12])
		f13 = faces[12]
	    faces[13] = (vz[12] + vz[11] + vz[7])
    	f14 = faces[13]
    	faces[14] = (vz[8] + vz[3] + vz[7])
    	f15 = faces[14]
    	faces[15] = (vz[10] + vz[3] + vz[8])
    	f16 = faces[15]
    	faces[16] = (vz[3] + vz[7] + vz[11])
    	f17 = faces[16]
    	faces[17] = (vz[4] + vz[3] + vz[10])
    	f18 = faces[17]
    	faces[18] = (vz[4] + vz[11] + vz[3])
    	f19 = faces[18]
    	faces[19] = (vz[6] + vz[11] + vz[4])
    	f20 = faces[19]
    
   		sort(faces)
   		
   		;face 1
    	if f1 <= $8d  {
    		draw_object(&f1_pairs,2)
    	}

		;face 2
		if f2 <= $8d {
			draw_object(&f2_pairs,2)	
		 }

		;face 3
		if f3 <= $8d  {
			draw_object(&f3_pairs,2)	
		 }

		;face 4			
		if f4 <= $8d  {		
			draw_object(&f4_pairs,2)	
 		}			
				
		;face 5	
		if f5 <= $8d {
			draw_object(&f5_pairs,2)
		}		
				
		;face 6			
		if f6 <= $8d {		
			draw_object(&f6_pairs,2)
		}		
				
		;face 7			
		if f7 <= $8d {		
			draw_object(&f7_pairs,2)
 		}		
				
		;face 8			
		if f8 <= $8d  {		
			draw_object(&f8_pairs,2)	
		}		
				
		;face 9			
		if f9 <= $8d  {		
			draw_object(&f9_pairs,2)
		}		
				
		;face 10			
		if f10 <= $8d  {		
			draw_object(&f10_pairs,2)	
		}			
				
		;face 11			
		if f11 <= $8d {			
			draw_object(&f11_pairs,2)	
 		}			
				
		;face 12			
		if f12 <= $8d {		
			draw_object(&f12_pairs,2)	
 		}			
 		
 		;face 13
    	if f13 <= $8d  {
    		draw_object(&f13_pairs,2)
    	}
 		
 		
		;face 14
		if f14 <= $8d {
			draw_object(&f14_pairs,2)	
		 }

		;face 15
		if f15 <= $8d  {
			draw_object(&f15_pairs,2)	
		 }

		;face 16			
		if f16 <= $8d  {		
			draw_object(&f16_pairs,2)	
 		}			
				
		;face 17
		if f17 <= $8d {
			draw_object(&f17_pairs,2)
		}		
				
		;face 18			
		if f18 <= $8d {		
			draw_object(&f18_pairs,2)
		}		
				
		;face 19			
		if f19 <= $8d {		
			draw_object(&f19_pairs,2)
 		}		
				
		;face 20			
		if f20 <= $8d  {		
			draw_object(&f20_pairs,2)	
		} 		
 		
 		
				
	}			
	else {							
		draw_object(&line_pairs,29)			
	}
    
  	
    						
	
	i_f = true
}

sub draw_bucky() {		; still a duplicate line here!
    ; drawing to back buffer (offscreen)   
    word[32] faces
    ubyte[] line_pairs = [2,4,4,32,32,56,56,54,54,30,30,2,2,34,34,58,
    					58,60,60,36,36,4,32,40,40,16,16,48,48,24,24,56,
    					40,44,44,36,60,28,28,52,52,20,20,44,20,8,
    					8,6,6,16,24,10,10,22,22,54,22,46,46,14,14,38,
    					38,30,14,5,5,7,7,18,18,42,42,38,42,34,18,50,
    					50,26,26,58,26,12,12,28,12,11,10,9,11,25,25,57,
    					57,59,59,27,27,11,57,33,33,41,41,17,17,49,49,25,
    					5,13,13,45,45,46,45,21,21,53,53,29,29,37,41,37,	
    					1,3,1,33,3,35,3,31,31,55,53,55,29,1,							
    					59,35,43,35,31,39,39,43,43,19,39,15,19,8,15,6,
    					15,47,51,19,51,27,47,23,55,23,37,13,7,17,9,21,			
    					9,23,19,43,47,48,49,50,51,52]	
    					
    															;near face outer
    ubyte[] f1_pairs = [7,5,5,14,14,38,38,42,42,18,18,7]		;12 o'clock 6
    ubyte[] f2_pairs = [14,46,46,22,22,54,54,30,30,38,38,14]	;2 o'clock 6
    ubyte[] f3_pairs = [22,54,54,56,56,24,24,10,10,22]			;3 o'clock 5
    ubyte[] f4_pairs = [24,56,56,32,32,40,40,16,16,48,48,24]	;4 o'clock 6
    ubyte[] f5_pairs = [40,16,16,6,6,8,8,20,20,44,44,40]		;6 o'clock 6
    ubyte[] f6_pairs = [44,20,20,52,52,28,28,60,60,36,36,44]	;7 o'clock 6
    ubyte[] f7_pairs = [60,28,28,12,12,26,26,58,58,60]			;9 o'clock 5
    ubyte[] f8_pairs = [26,50,50,18,18,42,42,34,34,58,58,26]	;10 o'clock 6
    															;near face inner
    ubyte[] f9_pairs = [42,38,38,30,30,2,2,34,34,42]			;12 o'clock 5
    ubyte[] f10_pairs = [30,54,54,56,56,32,32,4,4,2,2,30]		;3 o'clock 6
    ubyte[] f11_pairs = [4,32,32,40,40,44,44,36,36,4]			;6 o'clock 5
    ubyte[] f12_pairs = [2,4,4,36,36,60,60,58,58,34,34,2]		;9 o'clock 6
    
        														;far face outer
    ubyte[] f13_pairs = [5,7,7,17,17,41,41,37,37,13,13,5]		;12 o'clock 6
    ubyte[] f14_pairs = [17,49,49,25,25,57,57,33,33,41,41,17]	;2 o'clock 6
    
    ubyte[] f15_pairs = [25,11,11,27,27,59,59,57,57,25]			;3 o'clock 5
    
    ubyte[] f16_pairs = [27,51,51,19,19,43,43,35,35,59,59,27]	;4 o'clock 6
    ubyte[] f17_pairs = [19,8,8,6,6,15,15,39,39,43,43,19]		;6 o'clock 6
    ubyte[] f18_pairs = [39,15,15,47,47,23,23,55,55,31,31,39]	;7 o'clock 6
    
    ubyte[] f19_pairs = [55,23,23,9,9,21,21,53,53,55]			;9 o'clock 5
    
    ubyte[] f20_pairs = [53,21,21,45,45,13,13,37,37,29,29,53]	;10 o'clock 6
    															;far face inner
    ubyte[] f21_pairs = [37,41,41,33,33,1,1,29,29,37]			;12 o'clock 5
    ubyte[] f22_pairs = [1,33,33,57,57,59,59,35,35,3,3,1]		;3 o'clock 6
    ubyte[] f23_pairs = [3,35,35,43,43,39,39,31,31,3]			;6 o'clock 5
    ubyte[] f24_pairs = [31,55,55,53,53,29,29,1,1,3,3,31]		;9 o'clock 6
    
    ubyte[] f25_pairs = [9,10,10,22,22,46,46,45,45,21,21,9]
    ubyte[] f26_pairs = [9,10,10,24,24,48,48,47,47,23,23,9]
    ubyte[] f27_pairs = [11,12,12,26,26,50,50,49,49,25,25,11]
    ubyte[] f28_pairs = [11,12,12,28,28,52,52,51,51,27,27,11]
    
    ubyte[] f29_pairs = [5,13,13,45,45,46,46,14,14,5] 
    ubyte[] f30_pairs = [7,18,18,50,50,49,49,17,17,7]
    ubyte[] f31_pairs = [6,16,16,48,48,47,47,15,15,6]    
    ubyte[] f32_pairs = [8,20,20,52,52,51,51,19,19,8] 
    
    if hidden {
    
    	faces[0] = (vz[7] + vz[5] + vz[14] + vz[38] + vz[42] + vz[18])/6
		f1 = faces[0]
		
		faces[1] = (vz[14] + vz[46] + vz[22] + vz[54] + vz[30] + vz[38])/6
		f2 = faces[1]
		
		faces[2] = (vz[22] + vz[54] + vz[56] + vz[24] + vz[10])/5
		f3 = faces[2]
		
		faces[3] = (vz[24] + vz[56] + vz[32] + vz[40] + vz[16] + vz[48])/6
		f4 = faces[3]		
		
		faces[4] = (vz[40] + vz[16] + vz[6] + vz[8] + vz[20] + vz[44])/6
		f5 = faces[4]		
		
		faces[5] = (vz[44] + vz[20] + vz[52] + vz[28] + vz[60] + vz[36])/6
		f6 = faces[5]		
			
		faces[6] = (vz[60] + vz[28] + vz[12] + vz[26] + vz[58])/5
		f7 = faces[6]	
		
		faces[7] = (vz[26] + vz[50] + vz[18] + vz[42] + vz[34] + vz[58])/6
		f8 = faces[7]
		
		faces[8] = (vz[42] + vz[38] + vz[30] + vz[2] + vz[34])/5
		f9 = faces[8]
		
		faces[9] = (vz[30] + vz[54] + vz[56] + vz[32] + vz[4] + vz[2])/6
		f10 = faces[9]
		
		faces[10] = (vz[4] + vz[32] + vz[40] + vz[44] + vz[36])/5
		f11 = faces[10]		
		
		faces[11] = (vz[2] + vz[4] + vz[36] + vz[60] + vz[58] + vz[34])/6
		f12 = faces[11]		
			
		faces[12] = (vz[5] + vz[7] + vz[17] + vz[41] + vz[37] + vz[13])/6
		f13 = faces[12]	
		
		faces[13] = (vz[17] + vz[49] + vz[25] + vz[57] + vz[33] + vz[21])/6
		f14 = faces[13]		
		
		faces[14] = (vz[25] + vz[11] + vz[27] + vz[59] + vz[57])/5
		f15 = faces[14]		
			
		faces[15] = (vz[27] + vz[51] + vz[19] + vz[43] + vz[35] + vz[59])/6
		f16 = faces[15]			
				
		faces[16] = (vz[19] + vz[8] + vz[6] + vz[15] + vz[39] + vz[43])/6
		f17 = faces[16]				
		
		faces[17] = (vz[39] + vz[15] + vz[47] + vz[23] + vz[55] + vz[31])/6
		f18 = faces[17]		
		
		faces[18] = (vz[55] + vz[23] + vz[9] + vz[21] + vz[33])/5
		f19 = faces[18]		
			
		faces[19] = (vz[53] + vz[21] + vz[45] + vz[13] + vz[37] + vz[29])/6
		f20 = faces[19]			
		
		faces[20] = (vz[37] + vz[41] + vz[33] + vz[1] + vz[29])/5
		f21 = faces[20]		
			
		faces[21] = (vz[1] + vz[33] + vz[57] + vz[59] + vz[35] + vz[3])/6
		f22 = faces[21]			
		
		faces[22] = (vz[3] + vz[35] + vz[43] + vz[39] + vz[31])/5
		f23 = faces[22]		
		
		faces[23] = (vz[31] + vz[55] + vz[53] + vz[29] + vz[1] + vz[3])/6
		f24 = faces[23]		
		
		faces[24] = (vz[9] + vz[10] + vz[22] + vz[46] + vz[45] + vz[21])/6
		f25 = faces[24]	
		
		faces[25] = (vz[9] + vz[10] + vz[24] + vz[48] + vz[47] + vz[23])/6
		f26 = faces[25]		
				
			
		faces[26] = (vz[11] + vz[12] + vz[26] + vz[50] + vz[49] + vz[25])/6
		f27 = faces[26]				
					
		faces[27] = (vz[11] + vz[12] + vz[28] + vz[52] + vz[51] + vz[27])/6
		f28 = faces[27]						
						
		faces[28] = (vz[5] + vz[13] + vz[45] + vz[46] + vz[14])/5
		f29 = faces[28]						
						
		faces[29] = (vz[7] + vz[18] + vz[50] + vz[49] + vz[17])/5
		f30 = faces[29]							
								
		faces[30] = (vz[6] + vz[16] + vz[48] + vz[47] + vz[15])/5
		f31 = faces[30]								
								
		faces[31] = (vz[8] + vz[20] + vz[52] + vz[51] + vz[19])/5
		f32 = faces[31]									
									
		sort(faces)							
		
		;goto skip1
		
		;face 1
    	if f1 <= $2d  {
    		draw_object(&f1_pairs,5)   
   		}
    		
		;face 2
    	if f2 <= $2d  {
    		draw_object(&f2_pairs,5)   
    		}		
		
		;face 3
    	if f3 <= $2d  {
    		draw_object(&f3_pairs,4)	
    		}	
    		
		;face 4
    	if f4 <= $2d  {
    		draw_object(&f4_pairs,5)   
    		} 		
    		
		;face 5
    	if f5 <= $2d  {
    		draw_object(&f5_pairs,5)    
    		}		
    		
		;face 6
    	if f6 <= $2d  {
 		   	draw_object(&f6_pairs,5)    
    	}
    	
		;face 7
    	if f7 <= $2d  {
    		draw_object(&f7_pairs,4)    
    		}	
    		
		;face 8
    	if f8 <= $2d  {
    		draw_object(&f8_pairs,5)    
    		}		
    		
		;face 9
    	if f9 <= $2d  {
    		draw_object(&f9_pairs,4)    
    		}
    	
		;face 10
    	if f10 <= $2d  {
    		draw_object(&f10_pairs,5)    
    		}
    	
		;face 11
    	if f11 <= $2d  {
    		draw_object(&f11_pairs,4)    
    		}
    	
		;face 12
    	if f12 <= $2d  {
    		draw_object(&f12_pairs,5)    
    	}	
    ;goto skip		
skip1:	

		;face 13
    	if f13 <= $2d  {
    		draw_object(&f13_pairs,5)   
    		} 		
    		
		;face 14
    	if f14 <= $2d  {
    		draw_object(&f14_pairs,5)   
    		} 		
    		
		;face 15
    	if f15 <= $2d  {
    		draw_object(&f15_pairs,4)   
    		} 		
    		
		;face 16
    	if f16 <= $2d  {
    		draw_object(&f16_pairs,5)   
    	} 
    	
		;face 17
    	if f17 <= $2d  {
    		draw_object(&f17_pairs,5)    
    		}
    	
		;face 18
    	if f18 <= $2d  {
    		draw_object(&f18_pairs,5)   
    		} 	
    	
		;face 19
    	if f19 <= $2d  {
    		draw_object(&f19_pairs,4)   
    		} 
    	
		;face 20
    	if f20 <= $2d  {
    		draw_object(&f20_pairs,5)  
    		} 
    	
		;face 21
    	if f21 <= $2d  {
    		draw_object(&f21_pairs,4)   
    		} 
       	
		;face 22
    	if f22 <= $2d  {
    		draw_object(&f22_pairs,5)   
    		} 
    	
		;face 23
    	if f23 <= $2d  {
    		draw_object(&f23_pairs,4)    
    		}
    	
		;face 24
    	if f24 <= $2d  {
    		draw_object(&f24_pairs,5)   
    		} 

		;face 25
    	if f25 <= $2d  {
    		draw_object(&f25_pairs,5)   
    		} 

		;face 26
    	if f26 <= $2d  {
    		draw_object(&f26_pairs,5)   
    		} 

		;face 27
    	if f27 <= $2d  {
    		draw_object(&f27_pairs,5)   
    		}

		;face 28
    	if f28 <= $2d  {
    		draw_object(&f28_pairs,5)   
    		}

		;face 29
    	if f29 <= $2d  {
    		draw_object(&f29_pairs,4)   
    		}

		;face 30
    	if f30 <= $2d  {
    		draw_object(&f30_pairs,4)   
    		}

		;face 31
    	if f31 <= $2d  {
    		draw_object(&f31_pairs,4)   
    		}

		;face 32
    	if f32 <= $2d  {
    		draw_object(&f32_pairs,4)   
    		}

skip: 	
    } 
    else {				
		draw_object(&line_pairs,90) ;1 too many!
	}
	b_f = true
}

; draw a solid or just a face
sub draw_object(uword vertices, ubyte edges) {
	ubyte n
	ubyte q
	ubyte s
	ubyte e
	
	q = 0
	for n in 0 to edges {
		s = vertices[q]
		e = vertices[q+1]
		gfx3.line(vx[s],vy[s] as uword + 200,vx[e],vy[e] as uword + 200,1)
		q += 2
		}
}

	
sub init_buffers() {
        ; erase all vram
        cx16.VERA_CTRL = 0
        cx16.VERA_ADDR_L = 0
        cx16.VERA_ADDR_M = 0
        cx16.VERA_ADDR_H = %00010000        ; auto incr 1
        repeat $ffff
            cx16.VERA_DATA0 = 0
        repeat $f960
            cx16.VERA_DATA0 = 0
    }


;only copy what we need to, from back to front display buffer
sub blit2(uword x, uword y, uword width, ubyte height) {
	uword ix
	ubyte iy
	uword address = x + (y-200)*320

	for iy in 0 to height {
        cx16.VERA_CTRL = 0
        cx16.VERA_ADDR_L = lsb(address)
        cx16.VERA_ADDR_M = msb(address)
        cx16.VERA_ADDR_H = 0 | %00010000    
              
        cx16.VERA_CTRL = 1      
        cx16.VERA_ADDR_L = lsb(address)
        cx16.VERA_ADDR_M = msb(address)+ 250
        
        if y > 204 {
        	cx16.VERA_ADDR_H = 1 | %00010000 }     
        else {
        	cx16.VERA_ADDR_H = 0 | %00010000 }
        
        	for ix in 0 to width {
            	cx16.VERA_DATA0 = cx16.VERA_DATA1 }
            	
	address += 320
	}
}

sub vector(ubyte p1, ubyte p2, uword vec) {
	word vvx
	word vvy
	word vvz
	
		vvx = (vx[p2] - vx[p1]) as word
		vvy = (vy[p2] - vy[p1]) as word
		vvz = vz[p2] - vz[p1]
		
		cx16.memory_copy(&vvx,vec,2)
		cx16.memory_copy(&vvy,vec+2,2)
		cx16.memory_copy(&vvz,vec+4,2)
	}
	
sub cross_product(uword v1, uword v2, uword vres) {
	word v1x
	word v1y
	word v1z
	word v2x
	word v2y
	word v2z
	
	word vrx
	word vry
	word vrz
	
		cx16.memory_copy(v1,&v1x,2)
		cx16.memory_copy(v1+2,&v1y,2)
		cx16.memory_copy(v1+4,&v1z,2)
		
		cx16.memory_copy(v2,&v2x,2)
		cx16.memory_copy(v2+2,&v2y,2)
		cx16.memory_copy(v2+4,&v2z,2)
		
		vrx = v1y*v2z - v1z*v2y
		vry = v1z*v2x - v1x*v2z
		vrz = v1x*v2y - v1y*v2x
		
		cx16.memory_copy(&vrx,vres,2)
		cx16.memory_copy(&vry,vres+2,2)
		cx16.memory_copy(&vrz,vres+4,2)
		
}
		
		
	
	

	
	
sub getkey() {
	ubyte key = 0
	while key == 0 {
		key = c64.GETIN() }
	}
	
} ;end of main section

; copy of Prog8 gfx2 library
; modified by me to support 320x200 256c mode and double buffering (as mode 2)
; Bitmap pixel graphics routines for the CommanderX16
; Custom routines to use the full-screen 640x480 and 320x240 screen modes.
; (These modes are not supported by the documented GRAPH_xxxx kernal routines)
;
; No text layer is currently shown, text can be drawn as part of the bitmap itself.
; Note: for similar graphics routines that also work on the C-64, use the "graphics" module instead.
; Note: for color palette manipulation, use the "palette" module or write Vera registers yourself.
; Note: this library implements code for various resolutions and color depths. This takes up memory.
;       If you're memory constrained you should probably not use this built-in library,
;       but make a copy in your project only containing the code for the required resolution.
;
;
; SCREEN MODE LIST:
;   mode 0 = reset back to default text mode
;   mode 1 = bitmap 320 x 240 monochrome
;   mode 2 = bitmap 320 x 200 x 4c
;   mode 3 = bitmap 320 x 240 x 16c (TODO not yet implemented)
;   mode 4 = bitmap 320 x 240 x 256c  (like SCREEN $80 but using this api instead of kernal)
;   mode 5 = bitmap 640 x 480 monochrome
;   mode 6 = bitmap 640 x 480 x 4c
;   higher color dephts in highres are not supported due to lack of VRAM


gfx3 {

    ; read-only control variables:
    ubyte active_mode = 0
    uword width = 0
    uword height = 0
    ubyte bpp = 0
    bool monochrome_dont_stipple_flag = false            ; set to false to enable stippling mode in monochrome displaymodes

    sub screen_mode(ubyte mode) {
        when mode {
            1 -> {
                ; lores monochrome
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 64
                cx16.VERA_DC_VSCALE = 64
                cx16.VERA_L1_CONFIG = %00000100
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = 0
                width = 320
                height = 240
                bpp = 1
            }
            2 -> {
                ; 320x200 256c
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 64
                cx16.VERA_DC_VSCALE = 64
                cx16.VERA_L1_CONFIG = %00000111
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = 0
        		const ubyte vstart = 20
        		const ubyte vheight = 200
        		cx16.VERA_CTRL = %00000010
        		cx16.VERA_DC_VSTART = vstart
        		cx16.VERA_DC_VSTOP = vstart + vheight - 1                
                
                width = 320
                height = 200
                bpp = 8
            }            
            
            ; TODO modes 2, 3 not yet implemented
            4 -> {
                ; lores 256c
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 64
                cx16.VERA_DC_VSCALE = 64
                cx16.VERA_L1_CONFIG = %00000111
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = 0
                width = 320
                height = 240
                bpp = 8
            }
            5 -> {
                ; highres monochrome
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 128
                cx16.VERA_DC_VSCALE = 128
                cx16.VERA_L1_CONFIG = %00000100
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = %00000001
                width = 640
                height = 480
                bpp = 1
            }
            6 -> {
                ; highres 4c
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 128
                cx16.VERA_DC_VSCALE = 128
                cx16.VERA_L1_CONFIG = %00000101
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = %00000001
                width = 640
                height = 480
                bpp = 2
            }
            else -> {
                ; back to default text mode and colors
                cx16.VERA_CTRL = %10000000      ; reset VERA and palette
                c64.CINT()      ; back to text mode
                width = 0
                height = 0
                bpp = 0
                mode = 0
            }
        }

        active_mode = mode
        if bpp
            clear_screen()
    }
    
    sub clear_screen() {
        monochrome_stipple(false)
        position(0, 0)
        when active_mode {
            1 -> {
                ; lores monochrome
                repeat 240/2/8
                    cs_innerloop640()
            }
            ; TODO mode 2, 3
            2 -> {
                ; 320x200 256c
                repeat 200/2
                    cs_innerloop640()            
            }
            4 -> {
                ; lores 256c
                repeat 240/2
                    cs_innerloop640()
            }
            5 -> {
                ; highres monochrome
                repeat 480/8
                    cs_innerloop640()
            }
            6 -> {
                ; highres 4c
                repeat 480/4
                    cs_innerloop640()
            }
            ; modes 7 and 8 not supported due to lack of VRAM
        }
        position(0, 0)
    }


    sub monochrome_stipple(bool enable) {
        monochrome_dont_stipple_flag = not enable
    }

    sub rect(uword x, uword y, uword rwidth, uword rheight, ubyte color) {
        if rwidth==0 or rheight==0
            return
        horizontal_line(x, y, rwidth, color)
        if rheight==1
            return
        horizontal_line(x, y+rheight-1, rwidth, color)
        vertical_line(x, y+1, rheight-2, color)
        if rwidth==1
            return
        vertical_line(x+rwidth-1, y+1, rheight-2, color)
    }

    sub fillrect(uword x, uword y, uword rwidth, uword rheight, ubyte color) {
        if rwidth==0
            return
        repeat rheight {
            horizontal_line(x, y, rwidth, color)
            y++
        }
    }

    sub horizontal_line(uword x, uword y, uword length, ubyte color) {
        if length==0
            return
        when active_mode {
            1, 5 -> {
                ; monochrome modes, either resolution
                ubyte separate_pixels = (8-lsb(x)) & 7
                if separate_pixels as uword > length
                    separate_pixels = lsb(length)
                repeat separate_pixels {
                    ; TODO optimize this by writing a masked byte in 1 go
                    plot(x, y, color)
                    x++
                }
                length -= separate_pixels
                if length {
                    position(x, y)
                    separate_pixels = lsb(length) & 7
                    x += length & $fff8
                    %asm {{
                        lsr  length+1
                        ror  length
                        lsr  length+1
                        ror  length
                        lsr  length+1
                        ror  length
                        lda  color
                        bne  +
                        ldy  #0     ; black
                        bra  _loop
+                       lda  monochrome_dont_stipple_flag
                        beq  _stipple
                        ldy  #255       ; don't stipple
                        bra  _loop
_stipple                lda  y
                        and  #1         ; determine stipple pattern to use
                        bne  +
                        ldy  #%01010101
                        bra  _loop
+                       ldy  #%10101010
_loop                   lda  length
                        ora  length+1
                        beq  _done
                        sty  cx16.VERA_DATA0
                        lda  length
                        bne  +
                        dec  length+1
+                       dec  length
                        bra  _loop
_done
                    }}
                    repeat separate_pixels {
                        ; TODO optimize this by writing a masked byte in 1 go
                        plot(x, y, color)
                        x++
                    }
                }
                cx16.VERA_ADDR_H &= %00000111   ; vera auto-increment off again
            }
            2,4 -> {
                ; lores 256c
                position(x, y)
                %asm {{
                    lda  color
                    phx
                    ldx  length+1
                    beq  +
                    ldy  #0
-                   sta  cx16.VERA_DATA0
                    iny
                    bne  -
                    dex
                    bne  -
+                   ldy  length     ; remaining
                    beq  +
-                   sta  cx16.VERA_DATA0
                    dey
                    bne  -
+                   plx
                }}
            }
            6 -> {
                ; highres 4c
                ; TODO also mostly usable for lores 4c?
                color &= 3
                ubyte[4] colorbits
                ubyte ii
                for ii in 3 downto 0 {
                    colorbits[ii] = color
                    color <<= 2
                }
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    lda  cx16.VERA_ADDR_H
                    and  #%00000111         ; no auto advance
                    sta  cx16.VERA_ADDR_H
                    stz  cx16.VERA_CTRL     ; setup vera addr 0
                    lda  cx16.r1
                    and  #1
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0
                    sta  cx16.VERA_ADDR_L
                    lda  cx16.r0+1
                    sta  cx16.VERA_ADDR_M
                    phx
                    ldx  x
                }}

                repeat length {
                    %asm {{
                        txa
                        and  #3
                        tay
                        lda  cx16.VERA_DATA0
                        and  gfx3.plot.mask4c,y
                        ora  colorbits,y
                        sta  cx16.VERA_DATA0
                        cpy  #%00000011         ; next vera byte?
                        bne  ++
                        inc  cx16.VERA_ADDR_L
                        bne  ++
                        inc  cx16.VERA_ADDR_M
+                       bne  +
                        inc  cx16.VERA_ADDR_H
+                       inx                     ; next pixel
                    }}
                }

                %asm {{
                    plx
                }}
            }
        }
    }

    sub vertical_line(uword x, uword y, uword lheight, ubyte color) {
        when active_mode {
            1, 5 -> {
                ; monochrome, lo-res
                cx16.r15L = gfx3.plot.bits[x as ubyte & 7]           ; bitmask
                if color {
                    if monochrome_dont_stipple_flag {
                        ; draw continuous line.
                        position2(x,y,true)
                        if active_mode==1
                            set_both_strides(11)    ; 40 increment = 1 line in 320 px monochrome
                        else
                            set_both_strides(12)    ; 80 increment = 1 line in 640 px monochrome
                        repeat lheight {
                            %asm {{
                                lda  cx16.VERA_DATA0
                                ora  cx16.r15L
                                sta  cx16.VERA_DATA1
                            }}
                        }
                    } else {
                        ; draw stippled line.
                        if x&1 {
                            y++
                            lheight--
                        }
                        position2(x,y,true)
                        if active_mode==1
                            set_both_strides(12)    ; 80 increment = 2 line in 320 px monochrome
                        else
                            set_both_strides(13)    ; 160 increment = 2 line in 640 px monochrome
                        repeat lheight/2 {
                            %asm {{
                                lda  cx16.VERA_DATA0
                                ora  cx16.r15L
                                sta  cx16.VERA_DATA1
                            }}
                        }
                    }
                } else {
                    position2(x,y,true)
                    cx16.r15 = ~cx16.r15    ; erase pixels
                    if active_mode==1
                        set_both_strides(11)    ; 40 increment = 1 line in 320 px monochrome
                    else
                        set_both_strides(12)    ; 80 increment = 1 line in 640 px monochrome
                    repeat lheight {
                        %asm {{
                            lda  cx16.VERA_DATA0
                            and  cx16.r15L
                            sta  cx16.VERA_DATA1
                        }}
                    }
                }
            }
            2,4 -> {
                ; lores 256c
                ; set vera auto-increment to 320 pixel increment (=next line)
                position(x,y)
                cx16.VERA_ADDR_H = cx16.VERA_ADDR_H & %00000111 | (14<<4)
                %asm {{
                    ldy  lheight
                    beq  +
                    lda  color
-                   sta  cx16.VERA_DATA0
                    dey
                    bne  -
+
                }}
            }
            6 -> {
                ; highres 4c
                ; use TWO vera adress pointers simultaneously one for reading, one for writing, so auto-increment is possible
                if lheight==0
                    return
                position2(x,y,true)
                set_both_strides(13)    ; 160 increment = 1 line in 640 px 4c mode
                color &= 3
                color <<= gfx3.plot.shift4c[lsb(x) & 3]         ; TODO with lookup table
                ubyte @shared mask = gfx3.plot.mask4c[lsb(x) & 3]
                repeat lheight {
                    %asm {{
                        lda  cx16.VERA_DATA0
                        and  mask
                        ora  color
                        sta  cx16.VERA_DATA1
                    }}
                }
            }
        }

        sub set_both_strides(ubyte stride) {
            stride <<= 4
            cx16.VERA_CTRL = 0
            cx16.VERA_ADDR_H = cx16.VERA_ADDR_H & %00000111 | stride
            cx16.VERA_CTRL = 1
            cx16.VERA_ADDR_H = cx16.VERA_ADDR_H & %00000111 | stride
        }

    }

    sub line(uword @zp x1, uword @zp y1, uword @zp x2, uword @zp y2, ubyte color) {
        ; Bresenham algorithm.
        ; This code special-cases various quadrant loops to allow simple ++ and -- operations.
        if y1>y2 {
            ; make sure dy is always positive to have only 4 instead of 8 special cases
            cx16.r0 = x1
            x1 = x2
            x2 = cx16.r0
            cx16.r0 = y1
            y1 = y2
            y2 = cx16.r0
        }
        word @zp dx = (x2 as word)-x1
        word @zp dy = (y2 as word)-y1

        if dx==0 {
            vertical_line(x1, y1, abs(dy) as uword +1, color)
            return
        }
        if dy==0 {
            if x1>x2
                x1=x2
            horizontal_line(x1, y1, abs(dx) as uword +1, color)
            return
        }

        word @zp d = 0
        cx16.r13 = true      ; 'positive_ix'
        if dx < 0 {
            dx = -dx
            cx16.r13 = false
        }
        word @zp dx2 = dx*2
        word @zp dy2 = dy*2
        cx16.r14 = x1       ; internal plot X

        if dx >= dy {
            if cx16.r13 {
                repeat {
                    plot(cx16.r14, y1, color)
                    if cx16.r14==x2
                        return
                    cx16.r14++
                    d += dy2
                    if d > dx {
                        y1++
                        d -= dx2
                    }
                }
            } else {
                repeat {
                    plot(cx16.r14, y1, color)
                    if cx16.r14==x2
                        return
                    cx16.r14--
                    d += dy2
                    if d > dx {
                        y1++
                        d -= dx2
                    }
                }
            }
        }
        else {
            if cx16.r13 {
                repeat {
                    plot(cx16.r14, y1, color)
                    if y1 == y2
                        return
                    y1++
                    d += dx2
                    if d > dy {
                        cx16.r14++
                        d -= dy2
                    }
                }
            } else {
                repeat {
                    plot(cx16.r14, y1, color)
                    if y1 == y2
                        return
                    y1++
                    d += dx2
                    if d > dy {
                        cx16.r14--
                        d -= dy2
                    }
                }
            }
        }
    }

    sub circle(uword @zp xcenter, uword @zp ycenter, ubyte radius, ubyte color) {
        ; Midpoint algorithm.
        if radius==0
            return

        ubyte @zp xx = radius
        ubyte @zp yy = 0
        word @zp decisionOver2 = (1 as word)-xx
        ; R14 = internal plot X
        ; R15 = internal plot Y

        while xx>=yy {
            cx16.r14 = xcenter + xx
            cx16.r15 = ycenter + yy
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter + xx
            cx16.r15 = ycenter - yy
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter + yy
            cx16.r15 = ycenter + xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - yy
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter + yy
            cx16.r15 = ycenter - xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - yy
            plot(cx16.r14, cx16.r15, color)

            yy++
            if decisionOver2<=0
                decisionOver2 += (yy as word)*2+1
            else {
                xx--
                decisionOver2 += (yy as word -xx)*2+1
            }
        }
    }

    sub disc(uword @zp xcenter, uword @zp ycenter, ubyte @zp radius, ubyte color) {
        ; Midpoint algorithm, filled
        if radius==0
            return
        ubyte @zp yy = 0
        word @zp decisionOver2 = (1 as word)-radius

        while radius>=yy {
            horizontal_line(xcenter-radius, ycenter+yy, radius*$0002+1, color)
            horizontal_line(xcenter-radius, ycenter-yy, radius*$0002+1, color)
            horizontal_line(xcenter-yy, ycenter+radius, yy*$0002+1, color)
            horizontal_line(xcenter-yy, ycenter-radius, yy*$0002+1, color)
            yy++
            if decisionOver2<=0
                decisionOver2 += (yy as word)*2+1
            else {
                radius--
                decisionOver2 += (yy as word -radius)*2+1
            }
        }
    }

    sub plot(uword @zp x, uword y, ubyte color) {
        ubyte[8] @shared bits = [128, 64, 32, 16, 8, 4, 2, 1]
        ubyte[4] @shared mask4c = [%00111111, %11001111, %11110011, %11111100]
        ubyte[4] @shared shift4c = [6,4,2,0]

        when active_mode {
            1 -> {
                ; lores monochrome
                %asm {{
                    lda  x
                    eor  y
                    ora  monochrome_dont_stipple_flag
                    and  #1
                }}
                if_nz {
                    %asm {{
                        lda  x
                        and  #7
                        pha     ; xbits
                    }}
                    x /= 8
                    x += y*(320/8)
                    %asm {{
                        stz  cx16.VERA_CTRL
                        stz  cx16.VERA_ADDR_H
                        lda  x+1
                        sta  cx16.VERA_ADDR_M
                        lda  x
                        sta  cx16.VERA_ADDR_L
                        ply         ; xbits
                        lda  bits,y
                        ldy  color
                        beq  +
                        tsb  cx16.VERA_DATA0
                        bra  ++
+                       trb  cx16.VERA_DATA0
+
                    }}
                }
            }
            ; TODO mode 2,3
            2,4 -> {
                ; lores 256c
                void addr_mul_24_for_lores_256c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1
                    ora  #%00010000     ; enable auto-increment so next_pixel() can be used after this
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0+1
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0
                    sta  cx16.VERA_ADDR_L
                    lda  color
                    sta  cx16.VERA_DATA0
                }}
            }
            5 -> {
                ; highres monochrome
                %asm {{
                    lda  x
                    eor  y
                    ora  monochrome_dont_stipple_flag
                    and  #1
                }}
                if_nz {
                    %asm {{
                        lda  x
                        and  #7
                        pha     ; xbits
                    }}
                    x /= 8
                    x += y*(640/8)
                    %asm {{
                        stz  cx16.VERA_CTRL
                        stz  cx16.VERA_ADDR_H
                        lda  x+1
                        sta  cx16.VERA_ADDR_M
                        lda  x
                        sta  cx16.VERA_ADDR_L
                        ply     ; xbits
                        lda  bits,y
                        ldy  color
                        beq  +
                        tsb  cx16.VERA_DATA0
                        bra  ++
+                       trb  cx16.VERA_DATA0
+
                    }}
                }
            }
            6 -> {
                ; highres 4c
                ; TODO also mostly usable for lores 4c?
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                cx16.r2L = lsb(x) & 3       ; xbits
                color &= 3
                color <<= shift4c[cx16.r2L]     ; TODO with lookup table
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1L
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0H
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0L
                    sta  cx16.VERA_ADDR_L
                    ldy  cx16.r2L           ; xbits
                    lda  mask4c,y
                    and  cx16.VERA_DATA0
                    ora  color
                    sta  cx16.VERA_DATA0
                }}
            }
        }
    }

    sub pget(uword @zp x, uword y) -> ubyte {
        when active_mode {
            1 -> {
                ; lores monochrome
                %asm {{
                    lda  x
                    and  #7
                    pha     ; xbits
                }}
                x /= 8
                x += y*(320/8)
                %asm {{
                    stz  cx16.VERA_CTRL
                    stz  cx16.VERA_ADDR_H
                    lda  x+1
                    sta  cx16.VERA_ADDR_M
                    lda  x
                    sta  cx16.VERA_ADDR_L
                    ply         ; xbits
                    lda  plot.bits,y
                    and  cx16.VERA_DATA0
                    beq  +
                    lda  #1
+
                }}
            }
            ; TODO mode 2 and 3
            2,4 -> {
                ; lores 256c
                void addr_mul_24_for_lores_256c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0+1
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0
                    sta  cx16.VERA_ADDR_L
                    lda  cx16.VERA_DATA0
                }}
            }
            5 -> {
                ; hires monochrome
                %asm {{
                    lda  x
                    and  #7
                    pha     ; xbits
                }}
                x /= 8
                x += y*(640/8)
                %asm {{
                    stz  cx16.VERA_CTRL
                    stz  cx16.VERA_ADDR_H
                    lda  x+1
                    sta  cx16.VERA_ADDR_M
                    lda  x
                    sta  cx16.VERA_ADDR_L
                    ply     ; xbits
                    lda  plot.bits,y
                    and  cx16.VERA_DATA0
                    beq  +
                    lda  #1
+
                }}
            }
            6 -> {
                ; hires 4c
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1L
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0H
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0L
                    sta  cx16.VERA_ADDR_L
                    lda  cx16.VERA_DATA0
                    sta  cx16.r0L
                }}
                cx16.r1L = lsb(x) & 3
                cx16.r0L >>= gfx3.plot.shift4c[cx16.r1L]        ; TODO with lookup table
                return cx16.r0L & 3
            }
            else -> return 0
        }
    }

    sub position(uword @zp x, uword y) {
        ubyte bank
        when active_mode {
            1 -> {
                ; lores monochrome
                cx16.r0 = y*(320/8) + x/8
                cx16.vaddr(0, cx16.r0, 0, 1)
            }
            ; TODO modes 2,3
            2,4 -> {
                ; lores 256c
                void addr_mul_24_for_lores_256c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                bank = lsb(cx16.r1)
                cx16.vaddr(bank, cx16.r0, 0, 1)
            }
            5 -> {
                ; highres monochrome
                cx16.r0 = y*(640/8) + x/8
                cx16.vaddr(0, cx16.r0, 0, 1)
            }
            6 -> {
                ; highres 4c
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                bank = lsb(cx16.r1)
                cx16.vaddr(bank, cx16.r0, 0, 1)
            }
        }
    }

    sub position2(uword @zp x, uword y, bool also_port_1) {
        position(x, y)
        if also_port_1 {
            when active_mode {
                1, 5 -> cx16.vaddr(0, cx16.r0, 1, 1)
                ; TODO modes 2, 3
                2,4, 6 -> {
                    ubyte bank = lsb(cx16.r1)
                    cx16.vaddr(bank, cx16.r0, 1, 1)
                }
            }
        }
    }

    inline asmsub next_pixel(ubyte color @A) {
        ; -- sets the next pixel byte to the graphics chip.
        ;    for 8 bpp screens this will plot 1 pixel.
        ;    for 1 bpp screens it will plot 8 pixels at once (color = bit pattern).
        ;    for 2 bpp screens it will plot 4 pixels at once (color = bit pattern).
        %asm {{
            sta  cx16.VERA_DATA0
        }}
    }

    asmsub next_pixels(uword pixels @AY, uword amount @R0) clobbers(A, Y)  {
        ; -- sets the next bunch of pixels from a prepared array of bytes.
        ;    for 8 bpp screens this will plot 1 pixel per byte.
        ;    for 1 bpp screens it will plot 8 pixels at once (colors are the bit patterns per byte).
        ;    for 2 bpp screens it will plot 4 pixels at once (colors are the bit patterns per byte).
        %asm {{
            phx
            sta  P8ZP_SCRATCH_W1
            sty  P8ZP_SCRATCH_W1+1
            ldx  cx16.r0+1
            beq  +
            ldy  #0
-           lda  (P8ZP_SCRATCH_W1),y
            sta  cx16.VERA_DATA0
            iny
            bne  -
            inc  P8ZP_SCRATCH_W1+1       ; next page of 256 pixels
            dex
            bne  -

+           ldx  cx16.r0           ; remaining pixels
            beq  +
            ldy  #0
-           lda  (P8ZP_SCRATCH_W1),y
            sta  cx16.VERA_DATA0
            iny
            dex
            bne  -
+           plx
        }}
    }

    asmsub set_8_pixels_from_bits(ubyte bits @R0, ubyte oncolor @A, ubyte offcolor @Y) {
        ; this is only useful in 256 color mode where one pixel equals one byte value.
        %asm {{
            phx
            ldx  #8
-           asl  cx16.r0
            bcc  +
            sta  cx16.VERA_DATA0
            bra  ++
+           sty  cx16.VERA_DATA0
+           dex
            bne  -
            plx
            rts
        }}
    }

    const ubyte charset_bank = $1
    const uword charset_addr = $f000       ; in bank 1, so $1f000

    sub text_charset(ubyte charset) {
        ; -- select the text charset to use with the text() routine
        ;    the charset number is the same as for the cx16.screen_set_charset() ROM function.
        ;    1 = ISO charset, 2 = PETSCII uppercase+graphs, 3= PETSCII uppercase+lowercase.
        cx16.screen_set_charset(charset, 0)
    }

    sub text(uword @zp x, uword y, ubyte color, uword sctextptr) {
        ; -- Write some text at the given pixel position. The text string must be in screencode encoding (not petscii!).
        ;    You must also have called text_charset() first to select and prepare the character set to use.
        ;    NOTE: in monochrome (1bpp) screen modes, x position is currently constrained to multiples of 8 !  TODO allow per-pixel horizontal positioning
        ; TODO draw whole horizontal spans using vera auto increment if possible, instead of per-character columns
        uword chardataptr
        when active_mode {
            1, 5 -> {
                ; monochrome mode, either resolution
                cx16.r2 = 40
                if active_mode==5
                    cx16.r2 = 80
                while @(sctextptr) {
                    chardataptr = charset_addr + (@(sctextptr) as uword)*8
                    cx16.vaddr(charset_bank, chardataptr, 1, 1)
                    position(x,y)
                    %asm {{
                        lda  cx16.VERA_ADDR_H
                        and  #%111              ; don't auto-increment, we have to do that manually because of the ora
                        sta  cx16.VERA_ADDR_H
                        lda  color
                        sta  P8ZP_SCRATCH_B1
                        ldy  #8
-                       lda  P8ZP_SCRATCH_B1
                        bne  +                  ; white color, plot normally
                        lda  cx16.VERA_DATA1
                        eor  #255               ; black color, keep only the other pixels
                        and  cx16.VERA_DATA0
                        bra  ++
+                       lda  cx16.VERA_DATA0
                        ora  cx16.VERA_DATA1
+                       sta  cx16.VERA_DATA0
                        lda  cx16.VERA_ADDR_L
                        clc
                        adc  cx16.r2
                        sta  cx16.VERA_ADDR_L
                        bcc  +
                        inc  cx16.VERA_ADDR_M
+                       inc  x
                        bne  +
                        inc  x+1
+                       dey
                        bne  -
                    }}
                    sctextptr++
                }
            }
            4 -> {
                ; lores 256c
                while @(sctextptr) {
                    chardataptr = charset_addr + (@(sctextptr) as uword)*8
                    cx16.vaddr(charset_bank, chardataptr, 1, 1)
                    repeat 8 {
                        position(x,y)
                        y++
                        %asm {{
                            phx
                            ldx  color
                            lda  cx16.VERA_DATA1
                            sta  P8ZP_SCRATCH_B1
                            ldy  #8
-                           asl  P8ZP_SCRATCH_B1
                            bcc  +
                            stx  cx16.VERA_DATA0    ; write a pixel
                            bra  ++
+                           lda  cx16.VERA_DATA0    ; don't write a pixel, but do advance to the next address
+                           dey
                            bne  -
                            plx
                        }}
                    }
                    x+=8
                    y-=8
                    sctextptr++
                }
            }
            6 -> {
                ; hires 4c
                ; we're going to use a few cx16 registers to make sure every variable is in zeropage in the inner loop.
                cx16.r11L = color
                cx16.r8 = y
                while @(sctextptr) {
                    chardataptr = charset_addr + (@(sctextptr) as uword)*8
                    repeat 8 {
                        ; TODO rewrite this inner loop partly in assembly
                        ;      requires expanding the charbits to 2-bits per pixel (based on color)
                        ;      also it's way more efficient to draw whole horizontal spans instead of per-character
                        cx16.r9L = cx16.vpeek(charset_bank, chardataptr)  ; get the 8 horizontal character bits
                        cx16.r7 = x
                        repeat 8 {
                            cx16.r9L <<= 1
                            if_cs
                                plot(cx16.r7, cx16.r8, cx16.r11L)
                            cx16.r7++
                        }
                        chardataptr++
                        cx16.r8++
                    }
                    x+=8
                    cx16.r8-=8
                    sctextptr++
                }
            }
        }
    }

    asmsub cs_innerloop640() clobbers(Y) {
        %asm {{
            ldy  #80
-           stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            dey
            bne  -
            rts
        }}
    }

    asmsub addr_mul_24_for_highres_4c(uword yy @R2, uword xx @R3)  clobbers(A, Y) -> uword @R0, uword @R1 {
        ; yy * 160 + xx/4  (24 bits calculation)
        ; 24 bits result is in r0 and r1L (highest byte)
        %asm {{
            ldy  #5
-           asl  cx16.r2
            rol  cx16.r2+1
            dey
            bne  -
            lda  cx16.r2
            sta  cx16.r0
            lda  cx16.r2+1
            sta  cx16.r0+1
            asl  cx16.r0
            rol  cx16.r0+1
            asl  cx16.r0
            rol  cx16.r0+1

            ; xx >>= 2  (xx=R3)
            lsr  cx16.r3+1
            ror  cx16.r3
            lsr  cx16.r3+1
            ror  cx16.r3

            ; add r2 and xx (r3) to r0 (24-bits)
            stz  cx16.r1
            clc
            lda  cx16.r0
            adc  cx16.r2
            sta  cx16.r0
            lda  cx16.r0+1
            adc  cx16.r2+1
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+           clc
            lda  cx16.r0
            adc  cx16.r3
            sta  cx16.r0
            lda  cx16.r0+1
            adc  cx16.r3+1
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+
            rts
        }}
    }

    asmsub addr_mul_24_for_lores_256c(uword yy @R0, uword xx @AY) clobbers(A) -> uword @R0, ubyte @R1  {
        ; yy * 320 + xx (24 bits calculation)
        %asm {{
            sta  P8ZP_SCRATCH_W1
            sty  P8ZP_SCRATCH_W1+1
            lda  cx16.r0
            sta  P8ZP_SCRATCH_B1
            lda  cx16.r0+1
            sta  cx16.r1
            sta  P8ZP_SCRATCH_REG
            lda  cx16.r0
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            sta  cx16.r0
            lda  P8ZP_SCRATCH_B1
            clc
            adc  P8ZP_SCRATCH_REG
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+           ; now add the value to this 24-bits number
            lda  cx16.r0
            clc
            adc  P8ZP_SCRATCH_W1
            sta  cx16.r0
            lda  cx16.r0+1
            adc  P8ZP_SCRATCH_W1+1
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+           lda  cx16.r1
            rts
        }}
    }

}
